/*
**	hypervisor:vbr - VBR handling
*/
.section .gint.mapped, "ax"
.align 2

.global	_hypervisor_vbr_exch_reloc
.global	_hypervisor_vbr_tlbh_reloc
.global	_hypervisor_vbr_inth_reloc

/* hypervisor_vbr_exch_reloc(): Relocalized exception handler */
_hypervisor_vbr_exch_reloc:
	/* check if it's a hypervisor call */
	mov.l	exch_hypervisor_world_current, r4
	mov.l	exch_hypervisor_world, r3
	mov.l	@r4, r4
	mov.l	@r3, r3
	cmp/eq	r3, r4
	bf	exch_normal_behaviour

exch_trap_behaviour:
	/* check trapa exception and involve kernel ABI if needed */
	mov.l	exch_expevt_reg, r0
	mov.l	@r0, r0
	mov	#0x16, r1
	shll2	r1
	shll2	r1
	cmp/eq	r0, r1
	bf	exch_normal_behaviour
	mov.l	exch_kernel_abi, r0
	jmp	@r0
	nop

exch_normal_behaviour:
	/* normal behaviour */
	stc.l	spc, @-r4
	stc.l	ssr, @-r4

	/* get the "real" tlb handler address */
	mov.l	exch_hypervisor_vbr_redirect, r1
	mov.l	@r1, r1
	ldc	r1, vbr
	mov	#0x01, r2
	shll8	r2
	add	r2, r1

	/* jump into the "real" tlb handler and be sure that it will
	   return here with the same invoromnent */
	!stc	sr, r2
	mova	exch_jmp_over, r0
	ldc	r0, spc
	!ldc	r2, ssr
	jmp	@r1
	nop

.align 4
exch_jmp_over:
	/* get kernel stack and restore tlb event related information */
	mov.l	exch_hypervisor_world_current, r4
	mov.l	@r4, r4
	ldc.l	@r4+, ssr
	ldc.l	@r4+, spc

	/* force hypervisor VBR re-installation, because it possible that the
	   VBR has been changed during the handler */
	mov.l	exch_hypervisor_vbr_redirect, r0
	mov.l	exch_hypervisor_vbr, r1
	stc	vbr, r2
	mov.l	r2, @r0
	ldc.l	@r1+, vbr

	/* clean return */
	rte
	nop

.align 4
exch_expevt_reg:		.long	0xff000024
exch_kernel_abi:		.long	_hypervisor_kernel_abi
exch_hypervisor_world:		.long	_hypervisor_world
exch_hypervisor_world_current:	.long	_hypervisor_world_current
exch_hypervisor_vbr_redirect:	.long	_hypervisor_vbr_redirect
exch_hypervisor_vbr:		.long	_hypervisor_vbr
exch_end:
	.zero	0x300 - (exch_end - _hypervisor_vbr_exch_reloc)




/* hypervisor_vbr_tlbh_reloc(): Relocalized TLB handler */
_hypervisor_vbr_tlbh_reloc:
	/* get kernel stack and save tlb event related information */
	mov.l	tlbh_hypervisor_current_world, r4
	mov.l	@r4, r4
	stc.l	spc, @-r4
	stc.l	ssr, @-r4

	/* get the "real" tlb handler address */
	mov.l	tlbh_hypervisor_vbr_redirect, r1
	mov.l	@r1, r1
	ldc	r1, vbr
	mov	#0x04, r2
	shll8	r2
	add	r2, r1

	/* jump into the "real" interrupt handler and be sure that it will
	   return here with the same invoromnent */
	stc	sr, r2
	mova	tlbh_jmp_over, r0
	ldc	r0, spc
	ldc	r2, ssr
	jmp	@r1
	nop

.align 4
tlbh_jmp_over:
	/* get kernel stack and restore tlb event related information */
	mov.l	tlbh_hypervisor_current_world, r4
	mov.l	@r4, r4
	ldc.l	@r4+, ssr
	ldc.l	@r4+, spc

	/* force hypervisor VBR re-installation, because it possible that the
	   VBR has been changed during the handler */
	mov.l	tlbh_hypervisor_vbr_redirect, r0
	mov.l	tlbh_hypervisor_vbr, r1
	stc	vbr, r2
	mov.l	r2, @r0
	ldc.l	@r1+, vbr

	/* clean return */
	rte
	nop

.align 4
tlbh_hypervisor_vbr_redirect:	.long	_hypervisor_vbr_redirect
tlbh_hypervisor_vbr:		.long	_hypervisor_vbr
tlbh_hypervisor_current_world:	.long	_hypervisor_world_current
tlbh_end:
	.zero	0x200 - (tlbh_end - _hypervisor_vbr_tlbh_reloc)




/* hypervisor_vbr_inth_reloc(): Relocalized interrupt handler */
_hypervisor_vbr_inth_reloc:
	/* switch kernel stack */
	mov.l	inth_hypervisor_current_world, r4
	mov.l	@r4, r4
	mov.l	r15, @(4, r4)
	mov.l	@r4, r15

	/* save interrupt information */
	stc.l	gbr, @-r15
	sts.l	mach, @-r15
	sts.l	macl, @-r15
	stc.l	spc, @-r15
	stc.l	ssr, @-r15
	sts.l	pr, @-r15

	/* TODO: involve keyboard scanner */
	/* TODO: involved world switch to the hypervisor if needed */

	/* switch stack before calling the real handler */
	mov.l	r15, @r4
	mov.l	@(4, r4), r15

	/* get the "real" VBR interrupt handler address */
	mov.l	inth_hypervisor_vbr_redirect, r1
	mov.l	@r1, r1
	ldc	r1, vbr
	mov	#0x06, r2
	shll8	r2
	add	r2, r1

	/* jump into the "real" interrupt handler and be sure that it will
	   return here with the same invoromnent */
	stc	sr, r2
	mova	inth_jmp_over, r0
	ldc	r0, spc
	ldc	r2, ssr
	jmp	@r1
	nop

.align 4
inth_jmp_over:
	/* switch stack */
	mov.l	inth_hypervisor_current_world, r4
	mov.l	@r4, r4
	mov.l	r15, @(4, r4)
	mov.l	@r4, r15

	/* force hypervisor VBR re-installation, because it possible that the
	   VBR has been changed during the handler */
	mov.l	inth_hypervisor_vbr_redirect, r0
	mov.l	inth_hypervisor_vbr, r1
	stc	vbr, r2
	mov.l	r2, @r0
	ldc.l	@r1+, vbr

	/* restore interrupt information */
	lds.l	@r15+, pr
	ldc.l	@r15+, ssr
	ldc.l	@r15+, spc
	lds.l	@r15+, macl
	lds.l	@r15+, mach
	ldc.l	@r15+, gbr

	/* switch stack and exit */
	mov.l	r15, @r4
	mov.l	@(4, r4), r15

	/* Salut mon pote ! */
	rte
	nop

.align 4
inth_hypervisor_vbr:		.long	_hypervisor_vbr
inth_hypervisor_vbr_redirect:	.long	_hypervisor_vbr_redirect
inth_hypervisor_current_world:	.long	_hypervisor_world_current




/* define all symbols */
.section .gint.mappedrel, "aw"
.align 4
.global	_hypervisor_vbr_exch
.global	_hypervisor_vbr_tlbh
.global	_hypervisor_vbr_inth
_hypervisor_vbr_exch:	.long	_hypervisor_vbr_exch_reloc
_hypervisor_vbr_tlbh:	.long	_hypervisor_vbr_tlbh_reloc
_hypervisor_vbr_inth:	.long	_hypervisor_vbr_inth_reloc
